/*
#******[ Project information ]*******************
#
# Local Version      : 0.3
# Development SubVer : 09
# Started at         : 11 Nov 1997
# Last changes       : 31 Aug 2001
# Package Name       : lvwho : Subnet belonging check
#
#  Class      : shell utility
#  Description: Latvian subnets list checker
#  Platforms  : Linux,BSD,Solaris
#  Copyright  : (c) 1997-2001 SysMan
#  License    : GNU License
#  Author     : Alex Petrov <sysman@sysman.net>
#
#************************************************
 (c) 1997-2001 by Alex Petrov aka SysMan
  rebuild from old source tree Jul 2001.
  e-mail:sysman@sysman.net
  http://www.sysman.net/linux/lvwho/

# $Id: lvwho.c,v 1.9 2001/09/02 15:58:09 sysman Exp $
*/

#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <getopt.h>     /* long options */
#include <regex.h>

#include "lvwho.h"
#include "lvwho_vers.h"
/* ****************************
   *    Global/static vars    *
   **************************** */

static char *infile=INFILE;
static unsigned int line=0;     // cur. line number of infile

static unsigned short isp_n=0;             // number of isp's
static unsigned short snet_n=0;            // number of snet's

static struct {
char *name;                         // Pointers to ISP's names
unsigned short snet_start;           // start subnet
unsigned short snets;                // end subnet
} isp[MAX_ISPS];

static struct {
unsigned long  snet;             // Subnet
unsigned long  nmask;            // Mask
unsigned short p_isp;            // pointer->ISP[number] name
} snets[MAX_SUBNETS];           // Array of Snet's

/* 
  Values defined for Options, if nonzero - function work 
*/

static int  show_help=0;        /* display usage information and exit.    */
static int  show_version=0;     /* print the version and exit.            */
static int  debug=0;
static char *ip_opt_str=NULL;   /* pointer to option taken string ip */
static char *isp_opt_str=NULL;  /* pointer to option taken strin isp */

static struct option const long_opts[] =
{
 {"help",    no_argument,       &show_help,    'h'},
 {"version", no_argument,       &show_version, 'v'},
 {"file",    required_argument, NULL,          'f'},
 {"ip",      required_argument, NULL,          'i'},
 {"isp",     required_argument, NULL,          's'},
 {"debug",   required_argument, NULL,          'd'},
 {NULL, 0, NULL, 0}                               };
		  

/* ****************************
   *   Function definition    *
   **************************** */

int  read_ipacl(char *infile);
void add_isp(char *isp_name);
void add_snet(char *snet_str);
int  addr_in_snet(unsigned long addr,unsigned long snet,unsigned long mask);
int search_by_ip(void);
int search_by_isp(void);

/* ***********************************
   *  External functions definition. *
   *********************************** */

extern unsigned long str2addr(char *str);
extern char *addr2str(unsigned long addr);
extern int addr_in_snet(unsigned long addr,unsigned long snet,unsigned long mask);

/* version & help functions */
void version(void)
{
printf ("    [%s] Version %s\n",MY_PACKAGE,MY_VERSION_FULL );
printf ("--- %s Builded: %s ---\n",MY_PACKAGE_FULL,MY_BUILD_DATE );
printf ("--- COPYRIHT %s, %s ---\n",MY_COPYRIGHT, MY_LICENSE);
}

help(int status)
{
  if (status != 0)
    printf("Try `%s --help' for more information.\n",MY_PACKAGE);
  else
    {
 version();
printf(
"-------------------------------------------------------\n\
  -v, --version             version details.\n\
  -h, --help                help.\n\
  -f, --file <file>         input file (default=local.net)\n\
  -i, --ip <ip_address>     search subnet with include ip address.\n\
  -s, --isp <regex_pattern> search isp matched by regex pattern\n\
			    and list all assigned to it subnets.\n\
  -d, --debug <level>       turn on debug <level>\n");
printf("\nbugs & wishes send to Alex Petrov <sysman@sysman.net>.\n");
    }
  exit (1);
}

/* ******************************************************************** */
/*           **************      M A I N    *****************           */
/* ******************************************************************** */

int main(int argc, char **argv)
{
int i,c;          
int retc=3;

if(argc>1)
 {
 while ((c = getopt_long (argc, argv, "hvd:f:i:s:", long_opts, NULL)) != -1)
  {   switch (c)
   {
    case 0:
     break;
    case 'h':
      show_help=1;
     break;
    case 'v':
     show_version=1;
     break;
    case 'f':
     infile=optarg;
     break;
    case 'i':
     ip_opt_str=optarg;
     break;
    case 's':
     isp_opt_str=optarg;
     break;
    case 'd':
     debug = strtoul(optarg, NULL,10);
     printf("debug level  = %d\n",debug);
     break;

   default:
     help(1);
   }
  } /* end while */

  if (show_version) { version(); exit(0); }
  if (show_help)    help(0);
  
  if (ip_opt_str==0 && isp_opt_str==0) exit(0);
  
  read_ipacl(infile);        // Read subnets from file
  if(debug>1) printf("Readed from file: %d Subnets, %d ISPs\n",snet_n,isp_n);
  
  if (ip_opt_str!=0) search_by_ip();
  if (isp_opt_str!=0) search_by_isp();
  exit(retc);     // normal exit with code
  
 } // end of   if(argc>1)
  else help (0); /* if no args, show help and quit */
}
/* ****************************************************************** */
/*                      Procedures & Functions                        */
/* ****************************************************************** */


int read_ipacl(char *infile)
{
FILE *f;
char buf[128];
char *p,*arg;
char comment_start=0;

 if (!(f=fopen(infile,"r")))
  {
   fprintf(stderr,"Subnets file \"%s\" open error: %s\n",infile,strerror(errno));
   return 0;    // False
  }

 while ( fgets(buf,127,f) )
 {
  line++;
  p=buf;
   if ( (arg=strchr(p, '\n')) ) 
	    *arg='\0';                  /* clean out \n from str */
	    
  if(debug>4) printf("%3d: \"%s\"\n",line,buf);      // debuging stuff 
  
  while ((*p) && strchr(" \t", *p))     /* drop first spaces and tabs */
	    ++p;

  if ( strstr(p,"*/") && comment_start )
	{ comment_start=0;
	  continue; 
	}

  if ( comment_start )  
	  continue;                    /* Wait for END of comment */

  if ( strncmp(p,"/*",2)==0 ) 

	{ if( strstr(p,"*/") ) 
	       continue;                /* Skip one line comment */
	  comment_start=1;
	  continue; 
	}
	    
  if(strncmp(p,"###",3)==0) 
		continue;               /* Skip remarks */

  if(strncmp(p,"##",2)==0)              /* ##<name>:<site>:AS<numbers> */
    if(strchr(p+2,':')!=NULL)
		add_isp((p+2));         /* ADD new ISP header */
		
  if(strncasecmp(p,"isp:",4)==0)        /* ISP:<name>:<site>:AS<numbers> */
		add_isp(p+5);           /* ADD new ISP header */

  if( *p=='#' && isdigit(*(p+1)) )      /* #<subnet/mask> */
		add_snet((p+1));        /* ADD new subnet to current ISP */
		
  if ((!*p) || strchr(";:!", *p))
		continue;               // Skip Remarks 

  if ( strncmp(p,"//",2)==0 )
		continue;               // Skip Remarks   
 }

return 1;       // True, readed no-errors
}
/*   ------------------------------------
    ***  add_isp - add new isp to pool ***
     ------------------------------------
*/

void add_isp(char *isp_name)
{
int size;
char *p;

size=strlen(isp_name)+1;
 if ( (p=(char *)malloc(size)) == NULL ) 
   fprintf(stderr,"Memory alloc error: %s\n",strerror(errno)); 
 else 
 {
  memcpy(p,isp_name,size);
  isp[isp_n].name=p;
  isp[isp_n].snet_start=snet_n;
  isp_n++;
  if(debug>3) printf("ISP:[%s]\n",isp_name);
 }
}

/*       ------------------------------------------- 
    ***  add_snet - add subnet to current isp to pool ***
	 -------------------------------------------  
*/
void add_snet(char *snet_str)
{
char *mask,*ptr;
unsigned long tip,tcidr;

 mask=strchr(snet_str,'/');     /* divide string "subnet/mask_or_cidr" */
 *mask=0; mask++;

 ptr=snet_str+strlen(snet_str)-1;       /* drop out last spaces */
 while(*ptr==' ') { *ptr=0; ptr--; }

 ptr=mask+strlen(mask)-1;               /* drop out last spaces */
 while(*ptr==' ') { *ptr=0; ptr--; }
/* ** Parse over Subnet ** */

 tip=str2addr(snet_str);
 if( tip==0 ) 
   { 
    fprintf(stderr,"line %3d: wrong ip format \"%s\" - ignored\n",line,snet_str);
    return;
   }
 snets[snet_n].snet=tip;

/* ** Parse over MASK or CIDR ** */

 if(strchr(mask,'.')==0)
 {
   tcidr=strtoul(mask,0,10);
    if(tcidr<=32) tip=0xffffffff<<(32-tcidr);
    else  
     {
      fprintf(stderr,"line %3d: wrong CIDR \"%s\" - ignored\n",line,mask);
      return;
     }
 } else
 {
  if( (tip=str2addr(mask))==0);
   {
    fprintf(stderr,"line %3d: wrong mask \"%s\" - ignored\n",line,mask);
    return;
   }
 }
 snets[snet_n].nmask=tip;
 snets[snet_n].p_isp=(isp_n-1);
 isp[isp_n-1].snets++;
/* debug if need */
if(debug>3)  printf("Subnet [%s], mask/CIDR [%s/%d]\n",snet_str,mask,tcidr);
// if(debug>10) printf("hex subnet/mask %08x/%08x\n",snets[snet_n].snet,snets[snet_n].nmask);
      
 snet_n++;
}

/*   ----------------------------------------
    ***  search_by_ip - in which subnet(s) ***
     ----------------------------------------
*/
int search_by_ip()
{
int i;
int match=0;          /* quantity of matches */
unsigned long ip;
ip=str2addr(ip_opt_str);
 for(i=0;i<snet_n;i++)
  { if( addr_in_snet(ip,snets[i].snet,snets[i].nmask) )
    {
     if(match==0) printf("%s belonz to:\n",addr2str(ip));
     match++;
     printf("ISP:%s\n",isp[snets[i].p_isp].name);
     printf("#%s/",addr2str(snets[i].snet) );
     printf("#%s\n",addr2str(snets[i].nmask) );
    }
  }
if(match==0) printf("No matches found.\n");    
return(match);
}

/*   --------------------------------------------
    ***  search_by_isp - regex match ISP name  ***
     --------------------------------------------
*/
int search_by_isp()
{
regex_t reg_pat;
char reg_errb[101];
int i,s,z;
int match=0;

 i=regcomp(&reg_pat,isp_opt_str,REG_ICASE|REG_EXTENDED);
 if(i!=0)
  {
   regerror(i,&reg_pat,reg_errb,100);
   printf("Regex error: [%s]\n",reg_errb);
   regfree(&reg_pat);
   return 0;        // fail
  }
      
 for(i=0;i<isp_n;i++)
  { if( regexec(&reg_pat,isp[i].name,0,0,0)==0 )
    {
     if(match==0) printf("Match:\n");
     match++;
     printf("ISP:%s\n",isp[i].name);
      for(s=0;s<isp[i].snets;s++)
      { z=s+isp[i].snet_start;
       printf("#%s/",addr2str(snets[z].snet) );
       printf("#%s\n",addr2str(snets[z].nmask) );
      }
    }
  }
 regfree(&reg_pat);
 if(match==0) printf("No ISP matches found.\n");  
return(match);
}